@@ -17,11 +17,17 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *)
 open Type
+open Common
+
+type context_infos = {
+	com : Common.context;
+}
 
 type context = {
+	inf : context_infos;
 	ch : out_channel;
 	buf : Buffer.t;
-	path : module_path;
+	path : path;
 	mutable get_sets : (string * bool,string) Hashtbl.t;
 	mutable curclass : tclass;
 	mutable tabs : string;
@@ -36,7 +42,12 @@ type context = {
 	mutable constructor_block : bool;
 }
 
-let s_path ctx path p =
+let protect name =
+	match name with
+	| "Error" -> "_" ^ name
+	| _ -> name
+
+let s_path ctx stat path p =
 	match path with
 	| ([],name) ->
 		(match name with
@@ -48,31 +59,37 @@ let s_path ctx path p =
 		| _ -> name)
 	| (["flash"],"FlashXml__") ->
 		"Xml"
+	| (["flash"],"Error") ->
+		"Error"
+	| (["flash"],"Vector") ->
+		"Vector"
+	| (["flash";"xml"],"XML") ->
+		"XML"
+	| (["flash";"xml"],"XMLList") ->
+		"XMLList"
+	| (["haxe"],"Int32") when not stat ->
+		"int"
 	| (pack,name) ->
-		try
-			(match Hashtbl.find ctx.imports name with
-			| [p] when p = pack ->
-				name
-			| packs ->
-				if not (List.mem pack packs) then Hashtbl.replace ctx.imports name (pack :: packs);
-				Ast.s_type_path path)
-		with Not_found ->
-			Hashtbl.add ctx.imports name [pack];
-			name
+		let name = protect name in
+		let packs = (try Hashtbl.find ctx.imports name with Not_found -> []) in
+		if not (List.mem pack packs) then Hashtbl.replace ctx.imports name (pack :: packs);
+		Ast.s_type_path (pack,name)
+
+let reserved =
+	let h = Hashtbl.create 0 in
+	List.iter (fun l -> Hashtbl.add h l ())
+	(* these ones are defined in order to prevent recursion in some Std functions *)
+	["is";"as";"int";"uint";"const";"getTimer";"typeof";"parseInt";"parseFloat";
+	(* AS3 keywords which are not haXe ones *)
+	"each";"label";"finally";"with";"final";"internal";"native";"const";"namespace";"include";
+	(* we don't include get+set since they are not 'real' keywords, but they can't be used as method names *)
+	];
+	h
 
 let s_ident n =
-	match n with
-	| "is" -> "_is"
-	| "as" -> "_as"
-	| "int" -> "_int"
-	| "const" -> "_const"
-	| "getTimer" -> "_getTimer"
-	| "typeof" -> "_typeof"
-	| "parseInt" -> "_parseInt"
-	| "parseFloat" -> "_parseFloat"
-	| _ -> n
+	if Hashtbl.mem reserved n then "_" ^ n else n
 
-let init dir path =
+let init infos path =
 	let rec create acc = function
 		| [] -> ()
 		| d :: l ->
@@ -80,12 +97,13 @@ let init dir path =
 			if not (Sys.file_exists dir) then Unix.mkdir dir 0o755;
 			create (d :: acc) l
 	in
-	let dir = dir :: fst path in
+	let dir = infos.com.file :: fst path in
 	create [] dir;
 	let ch = open_out (String.concat "/" dir ^ "/" ^ snd path ^ ".as") in
 	let imports = Hashtbl.create 0 in
 	Hashtbl.add imports (snd path) [fst path];
 	{
+		inf = infos;
 		tabs = "";
 		ch = ch;
 		path = path;
@@ -134,12 +152,16 @@ let define_local ctx l =
 let spr ctx s = Buffer.add_string ctx.buf s
 let print ctx = Printf.kprintf (fun s -> Buffer.add_string ctx.buf s)
 
-let unsupported = Typer.error "This expression cannot be generated to AS3"
+let unsupported p = error "This expression cannot be generated to AS3" p
 
 let newline ctx =
-	match Buffer.nth ctx.buf (Buffer.length ctx.buf - 1) with
-	| '}' | '{' | ':' -> print ctx "\n%s" ctx.tabs
-	| _ -> print ctx ";\n%s" ctx.tabs
+	let rec loop p =
+		match Buffer.nth ctx.buf p with
+		| '}' | '{' | ':' -> print ctx "\n%s" ctx.tabs
+		| '\n' | '\t' -> loop (p - 1)
+		| _ -> print ctx ";\n%s" ctx.tabs
+	in
+	loop (Buffer.length ctx.buf - 1)
 
 let rec concat ctx s f = function
 	| [] -> ()
@@ -154,8 +176,6 @@ let open_block ctx =
 	ctx.tabs <- "\t" ^ ctx.tabs;
 	(fun() -> ctx.tabs <- oldt)
 
-let block = Transform.block
-
 let parent e =
 	match e.eexpr with
 	| TParenthesis _ -> e
@@ -172,9 +192,13 @@ let rec type_str ctx t p =
 			| "flash" :: _ , _ -> "String"
 			| _ -> "Object"
 		) else
-			s_path ctx e.e_path p
-	| TInst (c,_) ->
-		if (snd c.cl_path).[0] = '+' then "*" else s_path ctx c.cl_path p
+			s_path ctx true e.e_path p
+	| TInst ({ cl_path = ["flash"],"Vector" },[pt]) ->
+		"Vector.<" ^ type_str ctx pt p ^ ">"
+	| TInst (c,_) ->		
+		(match c.cl_kind with
+		| KNormal | KGeneric | KGenericInstance _ -> s_path ctx false c.cl_path p
+		| KTypeParameter | KExtension _ | KConstant _  -> "*")
 	| TFun _ ->
 		"Function"
 	| TMono r ->
@@ -225,27 +249,6 @@ let handle_break ctx e =
 
 let this ctx = if ctx.in_value <> None then "$this" else "this"
 
-let gen_function_header ctx name f params p =
-	let old = ctx.in_value in
-	let old_l = ctx.locals in
-	let old_li = ctx.inv_locals in
-	let old_t = ctx.local_types in
-	ctx.in_value <- None;
-	ctx.local_types <- List.map snd params @ ctx.local_types;
-	print ctx "function%s(" (match name with None -> "" | Some n -> " " ^ n);
-	concat ctx "," (fun (arg,o,t) ->
-		let arg = define_local ctx arg in
-		print ctx "%s : %s" arg (type_str ctx t p);
-		if o then spr ctx " = null";
-	) f.tf_args;
-	print ctx ") : %s " (type_str ctx f.tf_type p);
-	(fun () ->
-		ctx.in_value <- old;
-		ctx.locals <- old_l;
-		ctx.inv_locals <- old_li;
-		ctx.local_types <- old_t;
-	)
-
 let escape_bin s =
 	let b = Buffer.create 0 in
 	for i = 0 to String.length s - 1 do
@@ -264,7 +267,32 @@ let gen_constant ctx p = function
 	| TThis -> spr ctx (this ctx)
 	| TSuper -> spr ctx "super"
 
-let rec gen_call ctx e el =
+let gen_function_header ctx name f params p =
+	let old = ctx.in_value in
+	let old_l = ctx.locals in
+	let old_li = ctx.inv_locals in
+	let old_t = ctx.local_types in
+	ctx.in_value <- None;
+	ctx.local_types <- List.map snd params @ ctx.local_types;
+	print ctx "function%s(" (match name with None -> "" | Some n -> " " ^ n);
+	concat ctx "," (fun (arg,c,t) ->
+		let arg = define_local ctx arg in
+		print ctx "%s : %s" arg (type_str ctx t p);
+		match c with
+		| None -> ()
+		| Some c ->
+			spr ctx " = ";
+			gen_constant ctx p c
+	) f.tf_args;
+	print ctx ") : %s " (type_str ctx f.tf_type p);
+	(fun () ->
+		ctx.in_value <- old;
+		ctx.locals <- old_l;
+		ctx.inv_locals <- old_li;
+		ctx.local_types <- old_t;
+	)
+
+let rec gen_call ctx e el r =
 	match e.eexpr , el with
 	| TCall (x,_) , el ->
 		spr ctx "(";
@@ -312,6 +340,16 @@ let rec gen_call ctx e el =
 		gen_value ctx e;
 		print ctx ") %s.push(%s.substr(1))" ret tmp;
 		b();
+	| TLocal "__foreach__", [e] ->
+		let ret = (match ctx.in_value with None -> assert false | Some r -> r) in
+		print ctx "%s = new Array()" ret;
+		newline ctx;
+		let b = save_locals ctx in
+		let tmp = define_local ctx "$k" in
+		print ctx "for each(var %s : * in " tmp;
+		gen_value ctx e;
+		print ctx ") %s.push(%s)" ret tmp;
+		b();
 	| TLocal "__new__", e :: args ->
 		spr ctx "new ";
 		gen_value ctx e;
@@ -327,6 +365,29 @@ let rec gen_call ctx e el =
 		spr ctx ")";
 	| TLocal "__unprotect__", [e] ->
 		gen_value ctx e
+	| TLocal "__vector__", [e] ->
+		spr ctx (type_str ctx r e.epos);
+		spr ctx "(";
+		gen_value ctx e;
+		spr ctx ")"
+	| TField ({ eexpr = TTypeExpr (TClassDecl { cl_path = (["flash"],"Lib") }) },f), args ->
+		(match f, args with
+		| "vectorOfArray", [e] | "vectorConvert", [e] ->
+			(match follow r with
+			| TInst ({ cl_path = (["flash"],"Vector") },[t]) ->
+				print ctx "Vector.<%s>(" (type_str ctx t e.epos);
+				gen_value ctx e;
+				print ctx ")";
+			| _ -> assert false)
+		| "as", [e1;e2] ->
+			gen_value ctx e1;
+			spr ctx " as ";
+			gen_value ctx e2
+		| _ ->
+			gen_value ctx e;
+			spr ctx "(";
+			concat ctx "," (gen_value ctx) el;
+			spr ctx ")")
 	| _ ->
 		gen_value ctx e;
 		spr ctx "(";
@@ -357,6 +418,8 @@ and gen_field_access ctx t s =
 		| [], "String", "charCodeAt"
 		->
 			print ctx "[\"%s\"]" s
+		| [], "String", "cca" ->
+			print ctx ".charCodeAt"
 		| _ ->
 			print ctx ".%s" (s_ident s)
 	in
@@ -374,15 +437,15 @@ and gen_expr ctx e =
 	| TConst c ->
 		gen_constant ctx e.epos c
 	| TLocal s ->
-		spr ctx (try PMap.find s ctx.locals with Not_found -> Typer.error ("Unknown local " ^ s) e.epos)
+		spr ctx (try PMap.find s ctx.locals with Not_found -> error ("Unknown local " ^ s) e.epos)
 	| TEnumField (en,s) ->
-		print ctx "%s.%s" (s_path ctx en.e_path e.epos) (s_ident s)
+		print ctx "%s.%s" (s_path ctx true en.e_path e.epos) (s_ident s)
 	| TArray ({ eexpr = TLocal "__global__" },{ eexpr = TConst (TString s) }) ->
 		let path = (match List.rev (ExtString.String.nsplit s ".") with
 			| [] -> assert false
 			| x :: l -> List.rev l , x
 		) in
-		spr ctx (s_path ctx path e.epos)
+		spr ctx (s_path ctx false path e.epos)
 	| TArray (e1,e2) ->
 		gen_value ctx e1;
 		spr ctx "[";
@@ -403,7 +466,7 @@ and gen_expr ctx e =
    		gen_value ctx e;
 		gen_field_access ctx e.etype s
 	| TTypeExpr t ->
-		spr ctx (s_path ctx (t_path t) e.epos)
+		spr ctx (s_path ctx true (t_path t) e.epos)
 	| TParenthesis e ->
 		spr ctx "(";
 		gen_value ctx e;
@@ -432,11 +495,11 @@ and gen_expr ctx e =
 		let b = save_locals ctx in
 		print ctx "{";
 		let bend = open_block ctx in
-		let cb = (if not ctx.constructor_block then
+		let cb = (if not ctx.constructor_block || not (Codegen.constructor_side_effects e) then
 			(fun () -> ())
 		else begin
 			ctx.constructor_block <- false;
-			print ctx " if( !%s.skip_constructor ) {" (s_path ctx (["flash"],"Boot") e.epos);
+			print ctx " if( !%s.skip_constructor ) {" (s_path ctx true (["flash"],"Boot") e.epos);
             (fun() -> print ctx "}")
 		end) in
 		List.iter (fun e -> newline ctx; gen_expr ctx e) el;
@@ -449,11 +512,11 @@ and gen_expr ctx e =
 		let h = gen_function_header ctx None f [] e.epos in
 		let old = ctx.in_static in
 		ctx.in_static <- true;
-		gen_expr ctx (block f.tf_expr);
+		gen_expr ctx (mk_block f.tf_expr);
 		ctx.in_static <- old;
 		h();
-	| TCall (e,el) ->
-		gen_call ctx e el
+	| TCall (v,el) ->
+		gen_call ctx v el e.etype
 	| TArrayDecl el ->
 		spr ctx "[";
 		concat ctx "," (gen_value ctx) el;
@@ -475,7 +538,7 @@ and gen_expr ctx e =
 				gen_value ctx e
 		) vl;
 	| TNew (c,_,el) ->
-		print ctx "new %s(" (s_path ctx c.cl_path e.epos);
+		print ctx "new %s(" (s_path ctx true c.cl_path e.epos);
 		concat ctx "," (gen_value ctx) el;
 		spr ctx ")"
 	| TIf (cond,e,eelse) ->
@@ -511,7 +574,7 @@ and gen_expr ctx e =
 		handle_break();
 	| TObjectDecl fields ->
 		spr ctx "{ ";
-		concat ctx ", " (fun (f,e) -> print ctx "%s : " f; gen_value ctx e) fields;
+		concat ctx ", " (fun (f,e) -> print ctx "%s : " (s_ident f); gen_value ctx e) fields;
 		spr ctx "}"
 	| TFor (v,t,it,e) ->
 		let handle_break = handle_break ctx e in
@@ -530,13 +593,13 @@ and gen_expr ctx e =
 		handle_break();
 	| TTry (e,catchs) ->
 		spr ctx "try ";
-		gen_expr ctx (block e);
+		gen_expr ctx (mk_block e);
 		List.iter (fun (v,t,e) ->
 			newline ctx;
 			let b = save_locals ctx in
 			let v = define_local ctx v in
 			print ctx "catch( %s : %s )" v (type_str ctx t e.epos);
-			gen_expr ctx (block e);
+			gen_expr ctx (mk_block e);
 			b();
 		) catchs;
 	| TMatch (e,_,cases,def) ->
@@ -567,7 +630,7 @@ and gen_expr ctx e =
 						print ctx "%s : %s = %s.params[%d]" v (type_str ctx t e.epos) tmp n;
 					) l;
 					newline ctx);
-			gen_expr ctx (block e);
+			gen_expr ctx (mk_block e);
 			print ctx "break";
 			newline ctx;
 			b()
@@ -576,7 +639,7 @@ and gen_expr ctx e =
 		| None -> ()
 		| Some e ->
 			spr ctx "default:";
-			gen_expr ctx (block e);
+			gen_expr ctx (mk_block e);
 			print ctx "break";
 			newline ctx;
 		);
@@ -593,7 +656,7 @@ and gen_expr ctx e =
 				gen_value ctx e;
 				spr ctx ":";
 			) el;
-			gen_expr ctx (block e2);
+			gen_expr ctx (mk_block e2);
 			print ctx "break";
 			newline ctx;
 		) cases;
@@ -601,7 +664,7 @@ and gen_expr ctx e =
 		| None -> ()
 		| Some e ->
 			spr ctx "default:";
-			gen_expr ctx (block e);
+			gen_expr ctx (mk_block e);
 			print ctx "break";
 			newline ctx;
 		);
@@ -739,11 +802,11 @@ let generate_field ctx static f =
 	ctx.in_static <- static;
 	ctx.locals <- PMap.empty;
 	ctx.inv_locals <- PMap.empty;
-	let public = f.cf_public || Hashtbl.mem ctx.get_sets (f.cf_name,static) || (f.cf_name = "main" && static) || f.cf_name = "__resolve" in
+	let public = f.cf_public || Hashtbl.mem ctx.get_sets (f.cf_name,static) || (f.cf_name = "main" && static) || f.cf_name = "resolve" in
 	let rights = (if static then "static " else "") ^ (if public then "public" else "protected") in
 	let p = ctx.curclass.cl_pos in
 	match f.cf_expr with
-	| Some { eexpr = TFunction fd } when f.cf_set = MethodCantAccess ->
+	| Some { eexpr = TFunction fd } when f.cf_set = MethodCantAccess || f.cf_set = NeverAccess ->
 		print ctx "%s " rights;
 		let rec loop c =
 			match c.cl_super with
@@ -756,8 +819,9 @@ let generate_field ctx static f =
 		in
 		if not static then loop ctx.curclass;
 		let h = gen_function_header ctx (Some (s_ident f.cf_name)) fd f.cf_params p in
-		gen_expr ctx (block fd.tf_expr);
-		h()
+		gen_expr ctx (mk_block fd.tf_expr);
+		h();
+		newline ctx
 	| _ ->
 		if ctx.curclass.cl_path = (["flash"],"Boot") && f.cf_name = "init" then
 			generate_boot_init ctx
@@ -766,8 +830,13 @@ let generate_field ctx static f =
 			| TFun (args,r) ->
 				print ctx "function %s(" f.cf_name;
 				concat ctx "," (fun (arg,o,t) ->
-					print ctx "%s : %s" arg (type_str ctx t p);
-					if o then spr ctx " = null";
+					let tstr = type_str ctx t p in
+					print ctx "%s : %s" arg tstr;
+					if o then print ctx " = %s" (match tstr with
+						| "int" | "uint" -> "0"
+						| "Number" -> "NaN"
+						| "Boolean" -> "false"
+						| _ -> "null");					
 				) args;
 				print ctx ") : %s " (type_str ctx r p);
 			| _ -> ()
@@ -801,28 +870,35 @@ let generate_field ctx static f =
 				gen_value ctx e
 		end
 
-let define_getset ctx stat f =
-	let def name =
+let rec define_getset ctx stat c =
+	let def f name =
 		Hashtbl.add ctx.get_sets (name,stat) f.cf_name
 	in
-	(match f.cf_get with MethodAccess m -> def m | _ -> ());
-	(match f.cf_set with MethodAccess m -> def m | _ -> ())
+	let field f =
+		(match f.cf_get with MethodAccess m -> def f m | _ -> ());
+		(match f.cf_set with MethodAccess m -> def f m | _ -> ())
+	in
+	List.iter field (if stat then c.cl_ordered_statics else c.cl_ordered_fields);
+	match c.cl_super with
+	| Some (c,_) when not stat -> define_getset ctx stat c
+	| _ -> ()
+
 
 let generate_class ctx c =
 	ctx.curclass <- c;
-	List.iter (define_getset ctx false) c.cl_ordered_fields;
-	List.iter (define_getset ctx true) c.cl_ordered_statics;
+	define_getset ctx true c;
+	define_getset ctx false c;
 	ctx.local_types <- List.map snd c.cl_types;
 	let pack = open_block ctx in
-	print ctx "\tpublic %s%s %s " (match c.cl_dynamic with None -> "" | Some _ -> "dynamic ") (if c.cl_interface then "interface" else "class") (snd c.cl_path);
+	print ctx "\tpublic %s%s %s " (match c.cl_dynamic with None -> "" | Some _ -> if c.cl_interface then "" else "dynamic ") (if c.cl_interface then "interface" else "class") (snd c.cl_path);
 	(match c.cl_super with
 	| None -> ()
-	| Some (csup,_) -> print ctx "extends %s " (s_path ctx csup.cl_path c.cl_pos));
+	| Some (csup,_) -> print ctx "extends %s " (s_path ctx true csup.cl_path c.cl_pos));
 	(match c.cl_implements with
 	| [] -> ()
 	| l ->
 		spr ctx (if c.cl_interface then "extends " else "implements ");
-		concat ctx ", " (fun (i,_) -> print ctx "%s" (s_path ctx i.cl_path c.cl_pos)) l);
+		concat ctx ", " (fun (i,_) -> print ctx "%s" (s_path ctx true i.cl_path c.cl_pos)) l);
 	spr ctx "{";
 	let cl = open_block ctx in
 	(match c.cl_constructor with
@@ -849,7 +925,7 @@ let generate_class ctx c =
 let generate_main ctx c =
 	ctx.curclass <- c;
 	let pack = open_block ctx in
-	print ctx "\tpublic class __main__ extends %s {" (s_path ctx (["flash";"display"],"MovieClip") c.cl_pos);
+	print ctx "\tpublic class __main__ extends %s {" (s_path ctx true (["flash";"display"],"MovieClip") c.cl_pos);
 	let cl = open_block ctx in
 	newline ctx;
 	(match c.cl_ordered_statics with
@@ -857,7 +933,7 @@ let generate_main ctx c =
 		spr ctx "public function __main__() {";
 		let f = open_block ctx in
 		newline ctx;
-		print ctx "new %s(this)" (s_path ctx (["flash"],"Boot") c.cl_pos);
+		print ctx "new %s(this)" (s_path ctx true (["flash"],"Boot") c.cl_pos);
 		newline ctx;
 		gen_value ctx e;
 		f();
@@ -924,8 +1000,11 @@ let generate_base_enum ctx =
 	print ctx "}";
 	newline ctx
 
-let generate dir types =
-	let ctx = init dir ([],"enum") in
+let generate com =
+	let infos = {
+		com = com;
+	} in
+	let ctx = init infos ([],"enum") in
 	generate_base_enum ctx;
 	close ctx;
 	let boot = ref None in
@@ -935,7 +1014,7 @@ let generate dir types =
 		| TClassDecl c ->
 			let c = (match c.cl_path with
 				| ["flash"],"FlashXml__" -> { c with cl_path = [],"Xml" }
-				| _ -> c
+				| (pack,name) -> { c with cl_path = (pack,protect name) }
 			) in
 			(match c.cl_init with
 			| None -> ()
@@ -944,29 +1023,31 @@ let generate dir types =
 				()
 			else (match c.cl_path with
 			| [], "@Main" ->
-				let ctx = init dir ([],"__main__") in
+				let ctx = init infos ([],"__main__") in
 				generate_main ctx c;
 				close ctx;
 			| ["flash"], "Boot" ->
 				boot := Some c;
 			| _ ->
-				let ctx = init dir c.cl_path in
+				let ctx = init infos c.cl_path in
 				generate_class ctx c;
 				close ctx)
 		| TEnumDecl e ->
-			if e.e_extern then
+			let pack,name = e.e_path in
+			let e = { e with e_path = (pack,protect name) } in
+			if e.e_extern && e.e_path <> ([],"Void") then
 				()
 			else
-				let ctx = init dir e.e_path in
+				let ctx = init infos e.e_path in
 				generate_enum ctx e;
 				close ctx
 		| TTypeDecl t ->
 			()
-	) types;
+	) com.types;
 	match !boot with
 	| None -> assert false
 	| Some c ->
-		let ctx = init dir c.cl_path in
+		let ctx = init infos c.cl_path in
 		ctx.inits <- List.rev !inits;
 		generate_class ctx c;
 		close ctx
@@ -978,6 +1059,13 @@ let generate dir types =
    ---------------------------------------------------------------------------------------- *)
 open As3
 
+type access =
+	| APublic
+	| AProtected
+	| APrivate
+
+let cur_package = ref []
+
 let s_type_path = Ast.s_type_path
 
 let ident ctx p =
@@ -1019,10 +1107,14 @@ let rec real_type_path ctx p =
 		| A3MAttrib n ->
 			let path, name = loop n in
 			"$Attrib" :: path, name
+		| A3MParams (n,pl) ->
+			let t = type_path ctx n in
+			let params = "<" ^ (String.concat "," (List.map (fun t -> s_type_path (type_path ctx t)) pl)) ^ ">" in
+			fst t, (snd t ^ params)
 	in
 	loop (As3code.iget ctx.as3_names p)
 
-let type_path ctx p =
+and type_path ctx p =
 	match real_type_path ctx p with
 	| [] , "Object" -> [] , "Dynamic"
 	| [] , "Boolean" -> [] , "Bool"
@@ -1033,6 +1125,9 @@ let type_path ctx p =
 	| [] , "void" -> [] , "Void"
 	| [] , "Function" -> [] , "Dynamic"
 	| [] , "Class" -> [] , "Class<Dynamic>"
+	| [] , "Error" -> ["flash"], "Error"
+	| ["__AS3__";"vec"] , "Vector" -> ["flash"], "Vector"
+	| pack, cl when pack = !cur_package -> [], cl
 	| path -> path
 
 let ident_rights ctx id =
@@ -1040,11 +1135,11 @@ let ident_rights ctx id =
 	| A3MName (id,r) ->
 		let name = ident ctx id in
 		(match As3code.iget ctx.as3_namespaces r with
-		| A3NNamespace i when As3code.iget ctx.as3_idents i = "http://www.adobe.com/2006/flex/mx/internal" -> false, false, "$" ^ name
-		| A3NPublic _ | A3NNamespace _ -> false , false, name
-		| A3NProtected _ -> false, true, name
-		| _ -> true , false, name)
-	| _ -> false, false, "???"
+		| A3NNamespace i when As3code.iget ctx.as3_idents i = "http://www.adobe.com/2006/flex/mx/internal" -> APublic, "$" ^ name
+		| A3NPublic _ | A3NNamespace _ -> APublic, name
+		| A3NProtected _ -> AProtected, name
+		| _ -> APrivate, name)
+	| _ -> APublic, "???"
 
 let rec create_dir acc = function
 	| [] -> ()
@@ -1109,15 +1204,40 @@ let gen_method ctx ch name mt =
 		incr p;
 		(if opt_val <> None then "?" else "") ^ name ^ " : " ^ t
 	) m.mt3_args in
-	let vargs = if m.mt3_var_args then " /* ...arguments */" else "" in
+	let vargs = if m.mt3_var_args then
+		(if m.mt3_args = [] then "" else ",") ^ " ?p1 : Dynamic, ?p2 : Dynamic, ?p3 : Dynamic, ?p4 : Dynamic, ?p5 : Dynamic "
+	else
+		""
+	in
 	IO.printf ch "function %s(%s%s) : %s;\n" name (String.concat ", " params) vargs ret
 
-let gen_fields ctx ch fields stat =
-	let fields = List.sort (fun f1 f2 -> compare (ident_rights ctx f1.f3_name) (ident_rights ctx f2.f3_name)) (Array.to_list fields) in
+let is_fun = function
+	| A3FMethod m -> m.m3_kind = MK3Normal
+	| _ -> false
+
+let sort_fields ctx f1 f2 =
+	let acc1, name1 = ident_rights ctx f1.f3_name in
+	let acc2, name2 = ident_rights ctx f2.f3_name in
+	let fun1 = is_fun f1.f3_kind in
+	let fun2 = is_fun f2.f3_kind in
+	compare (acc1,fun1,name1) (acc2,fun2,name2)
+
+let gen_fields ctx ch fields stat construct =
+	let fields = List.sort (sort_fields ctx) (Array.to_list fields) in
+	let construct = ref construct in
+	let gen_construct() =
+		match !construct with
+		| None -> ()
+		| Some c ->
+			construct := None;
+			IO.printf ch "\t";
+			gen_method ctx ch "new" c;
+	in
 	List.iter (fun f ->
-		let priv, prot, name = ident_rights ctx f.f3_name in
-		let rights = (if priv then "//" else "") ^ (if priv || prot then "private " else "") ^ (if stat then "static " else "") in
-		if name.[0] = '$' then
+		let acc, name = ident_rights ctx f.f3_name in
+		let rights = (match acc with APrivate -> "//private " | AProtected -> "private " | APublic -> "") ^ (if stat then "static " else "") in
+		if acc <> APublic || is_fun f.f3_kind then gen_construct();
+		if name.[0] = '$' || acc = APrivate then
 			()
 		else match f.f3_kind with
 		| A3FMethod m ->
@@ -1127,7 +1247,7 @@ let gen_fields ctx ch fields stat =
 			(match m.m3_kind with
 			| MK3Normal ->
 				IO.printf ch "\t%s" rights;
-				gen_method ctx ch name m.m3_type				
+				gen_method ctx ch name m.m3_type
 			| MK3Getter ->
 				let set = has_getset fields f m in
 				let set_str = if set then "" else "(default,null)" in
@@ -1149,11 +1269,14 @@ let gen_fields ctx ch fields stat =
 			assert false
 		| A3FClass _ ->
 			IO.printf ch "\t// ????\n"
-	) fields
+	) fields;
+	gen_construct()
 
 let genhx_class ctx c s =
 	let base_path = "hxclasses" in
+	cur_package := [];
 	let pack , name = real_type_path ctx c.cl3_name in
+	cur_package := pack;
 	let skip = (match pack with
 		| [_;x] when String.length x > 3 && String.sub x 0 3 = "as$" -> true
 		| _ when name.[0] = '_' -> true
@@ -1167,7 +1290,25 @@ let genhx_class ctx c s =
 	let f = open_out (base_path ^ "/" ^ (match pack with [] -> "" | l -> String.concat "/" l ^ "/") ^ name ^ ".hx") in
 	let ch = IO.output_channel f in
 	if pack <> [] then IO.printf ch "package %s;\n\n" (String.concat "." pack);
-	IO.printf ch "extern %s %s" (if c.cl3_interface then "interface" else "class") name;
+	let enum_fields, isenum = (try
+		if Array.length c.cl3_fields > 0 || c.cl3_interface || Array.length s.st3_fields = 0 then raise Exit;
+		(match c.cl3_super with None -> () | Some p -> if type_path ctx p <> ([],"Dynamic") then raise Exit);
+		let etype = ref None in
+		let fields = List.map (fun f ->
+			(match f.f3_kind with
+			| A3FVar v ->
+				let t = type_val ctx v.v3_type (Some v.v3_value) in
+				(match !etype with
+				| None -> etype := Some t
+				| Some t2 -> if t <> t2 then raise Exit);
+			| _ -> raise Exit);
+			let prot, name = ident_rights ctx f.f3_name in
+			if prot <> APublic then raise Exit;
+			name
+		) (Array.to_list s.st3_fields) in
+		fields, true
+	with Exit -> [], false) in
+	IO.printf ch "extern %s %s" (if isenum then "enum" else if c.cl3_interface then "interface" else "class") name;
 	let prev = ref (match c.cl3_super with
 	| None -> false
 	| Some p ->
@@ -1183,16 +1324,19 @@ let genhx_class ctx c s =
 		IO.printf ch " implements %s" (s_type_path (type_path ctx i));
 	) c.cl3_implements;
 	IO.printf ch " {\n";
-	IO.printf ch "\t";
-	if not c.cl3_interface then gen_method ctx ch "new" c.cl3_construct;
-	gen_fields ctx ch c.cl3_fields false;
-	gen_fields ctx ch s.st3_fields true;
+	if isenum then
+		List.iter (fun f -> IO.printf ch "\t%s;\n" f) (List.sort compare enum_fields)
+	else begin
+		let construct = (if not c.cl3_interface && Array.length c.cl3_fields > 0 then Some c.cl3_construct else None) in
+		gen_fields ctx ch c.cl3_fields false construct;
+		gen_fields ctx ch s.st3_fields true None;
+	end;
 	IO.printf ch "}\n";
 	prerr_endline ";";
 	IO.close_out ch
 
-let genhx file =
-	let file = (try Plugin.find_file file with Not_found -> failwith ("File not found : " ^ file)) in
+let genhx com =
+	let file = (try Common.find_file com com.file with Not_found -> failwith ("File not found : " ^ com.file)) in
 	let ch = IO.input_channel (open_in_bin file) in
 	SwfParser.full_parsing := true;
 	let _, swf = Swf.parse ch in
