@@ -1,6 +1,6 @@
 (*
  *  Haxe Compiler
- *  Copyright (c)2005 Nicolas Cannasse
+ *  Copyright (c)2005-2008 Nicolas Cannasse
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -18,26 +18,17 @@
  *)
 open Printf
 open Genswf
+open Common
 
-type target =
-	| No
-	| Js of string
-	| Swf of string
-	| Neko of string
-	| As3 of string
+let version = 203
 
 let prompt = ref false
-let alt_format = ref false
-let has_error = ref false
-let auto_xml = ref false
 let display = ref false
+let measure_times = ref false
 
 let executable_path() =
 	Extc.executable_path()
 
-let get_full_path file =
-	Extc.get_full_path file
-
 let normalize_path p =
 	let l = String.length p in
 	if l = 0 then
@@ -46,16 +37,11 @@ let normalize_path p =
 		| '\\' | '/' -> p
 		| _ -> p ^ "/"
 
-let warn msg p =
+let message msg p =
 	if p = Ast.null_pos then
 		prerr_endline msg
 	else begin
-		let error_printer file line =
-			if !alt_format then
-				sprintf "%s(%d):" file line
-			else
-				sprintf "%s:%d:" file line
-		in
+		let error_printer file line = sprintf "%s:%d:" file line in
 		let epos = Lexer.get_error_pos error_printer p in
 		let msg = String.concat ("\n" ^ epos ^ " : ") (ExtString.String.nsplit msg "\n") in
 		prerr_endline (sprintf "%s : %s" epos msg)
@@ -69,7 +55,7 @@ let do_exit() =
 	exit 1
 
 let report msg p =
-	warn msg p;
+	message msg p;
 	do_exit()
 
 let htmlescape s =
@@ -84,15 +70,6 @@ let report_list l =
 	) (List.sort (fun (a,_,_) (b,_,_) -> compare a b) l);
 	prerr_endline "</list>"
 
-let type_error e p =
-	warn (Typer.error_msg e) p;
-	has_error := true
-
-let parse_error e p =
-	Lexer.save_lines();
-	warn (Parser.error_msg e) p;
-	has_error := true
-
 let file_extension f =
 	let cl = ExtString.String.nsplit f "." in
 	match List.rev cl with
@@ -124,10 +101,19 @@ let make_path f =
 	in
 	loop cl
 
-let read_type_path p cp =
+let read_type_path com p =
 	let classes = ref [] in
 	let packages = ref [] in
-	let p = (match p with "flash" :: l when Plugin.defined "flash9" -> "flash9" :: l | _ -> p) in
+	let p = (match p with
+		| x :: l ->
+			(try
+				match PMap.find x com.package_rules with
+				| Directory d -> d :: l
+				| _ -> p
+			with
+				Not_found -> p)
+		| _ -> p
+	) in
 	List.iter (fun path ->
 		let dir = path ^ String.concat "/" p in
 		let r = (try Sys.readdir dir with _ -> [||]) in
@@ -139,23 +125,26 @@ let read_type_path p cp =
 				if String.length c < 2 || String.sub c (String.length c - 2) 2 <> "__" then classes := c :: !classes;
 			end;
 		) r;
-	) cp;
+	) com.class_path;
 	List.sort compare (!packages), List.sort compare (!classes)
 
 let delete_file f = try Sys.remove f with _ -> ()
 
+let expand_env path =
+	let r = Str.regexp "%\\([^%]+\\)%" in
+	Str.global_substitute r (fun s -> try Sys.getenv (Str.matched_group 1 s) with Not_found -> "") path
+
 let parse_hxml file =
-	let ch = (try open_in file with _ -> failwith ("File not found " ^ file)) in
-	let lines = Std.input_list ch in
-	close_in ch;
+	let ch = IO.input_channel (try open_in_bin file with _ -> failwith ("File not found " ^ file)) in
+	let lines = Str.split (Str.regexp "[\r\n]+") (IO.read_all ch) in
+	IO.close_in ch;
 	List.concat (List.map (fun l ->
 		let l = ExtString.String.strip l in
-		(*// disabled - need additional str.cmxa linkage
-			let renv = Str.regexp "%\\([A-Za-z0-9_]+\\)%" in
-			let l = Str.global_substitute renv (fun _ ->
+		let renv = Str.regexp "%\\([A-Za-z0-9_]+\\)%" in
+		let l = Str.global_substitute renv (fun _ ->
 			let e = Str.matched_group 1 l in
 			try Sys.getenv e with Not_found -> "%" ^ e ^ "%"
-		) l in  *)
+		) l in
 		if l = "" || l.[0] = '#' then
 			[]
 		else if l.[0] = '-' then
@@ -168,9 +157,6 @@ let parse_hxml file =
 			[l]
 	) lines)
 
-
-let base_defines = !Plugin.defines
-
 exception Hxml_found
 
 let rec process_params acc = function
@@ -183,30 +169,35 @@ let rec process_params acc = function
 		process_params (x :: acc) l
 
 and init params =
-try
-	let version = 119 in
-	let version_str = Printf.sprintf "%d.%.2d" (version / 100) (version mod 100) in
-	let usage = "Haxe Compiler " ^ version_str ^ " - (c)2005-2008 Motion-Twin\n Usage : haxe.exe [options] <class names...>\n Options :" in
+	let usage = Printf.sprintf
+		"Haxe Compiler %d.%.2d - (c)2005-2009 Motion-Twin\n Usage : haxe.exe %s <class names...>\n Options :"
+		(version / 100) (version mod 100) (if Sys.os_type = "Win32" then ".exe" else "")
+	in
 	let classes = ref [([],"Std")] in
-	let target = ref No in
+	let com = Common.create version in
+try
 	let xml_out = ref None in
-	let main_class = ref None in
-	let swf_infos = {
-		swf_version = 8;
-		swf_header = None;
-		swf_lib = None;
-	} in
-	let hres = Hashtbl.create 0 in
+	let swf_header = ref None in
+	let swf_lib = ref None in
 	let cmds = ref [] in
 	let excludes = ref [] in
 	let libs = ref [] in
-	let gen_hx = ref false in
+	let has_error = ref false in
+	let gen_as3 = ref false in
 	let no_output = ref false in
-	Plugin.defines := base_defines;
-	Plugin.define ("haxe_" ^ string_of_int version);
-	Typer.check_override := false;
-	Typer.forbidden_packages := ["js"; "neko"; "flash"];
-	Parser.display_error := parse_error;
+	let did_something = ref false in
+	let root_packages = ["neko"; "flash"; "flash9"; "js"; "php"; "cpp"] in
+	Common.define com ("haxe_" ^ string_of_int version);
+	com.warning <- message;
+	com.error <- (fun msg p ->
+		message msg p;
+		has_error := true;
+	);
+	Parser.display_error := (fun e p ->
+		Lexer.save_lines();
+		message (Parser.error_msg e) p;
+		has_error := true;
+	);
 	Parser.use_doc := false;
 	(try
 		let p = Sys.getenv "HAXE_LIBRARY_PATH" in
@@ -219,45 +210,87 @@ try
 			| l ->
 				l
 		in
-		Plugin.class_path := List.map normalize_path (loop (ExtString.String.nsplit p ":"))
+		let parts = "" :: Str.split_delim (Str.regexp "[;:]") p in
+		com.class_path <- List.map normalize_path (loop parts)
 	with
 		Not_found ->
 			if Sys.os_type = "Unix" then
-				Plugin.class_path := ["/usr/lib/haxe/std/";"/usr/local/lib/haxe/std/";"";"/"]
+				com.class_path <- ["/usr/lib/haxe/std/";"/usr/local/lib/haxe/std/";"";"/"]
 			else
 				let base_path = normalize_path (try executable_path() with _ -> "./") in
-				Plugin.class_path := [base_path ^ "std/";"";"/"]);
-	let check_targets() =
-		if !target <> No then failwith "Multiple targets";
+				com.class_path <- [base_path ^ "std/";"";"/"]);
+	let set_platform pf name file =
+		if com.platform <> Cross then failwith "Multiple targets";
+		com.platform <- pf;
+		com.file <- file;
+		let forbid acc p = if p = name || PMap.mem p acc then acc else PMap.add p Forbidden acc in
+		com.package_rules <- List.fold_left forbid com.package_rules root_packages;
+		Common.define com name; (* define platform name *)
+		Unix.putenv "__file__" file;
+		Unix.putenv "__platform__" file;
+		if (pf = Flash || pf = Flash9) && file_extension file = "swc" then Common.define com "swc";
 	in
-	let define f = Arg.Unit (fun () -> Plugin.define f) in
-	let args_spec = [
+	let define f = Arg.Unit (fun () -> Common.define com f) in
+	let basic_args_spec = [
 		("-cp",Arg.String (fun path ->
-			Plugin.class_path := normalize_path path :: !Plugin.class_path
+			com.class_path <- normalize_path path :: com.class_path
 		),"<path> : add a directory to find source files");
-		("-js",Arg.String (fun file ->
-			check_targets();
-			Typer.forbidden_packages := ["neko"; "flash"];
-			target := Js file
-		),"<file> : compile code to JavaScript file");
+		("-js",Arg.String (set_platform Js "js"),"<file> : compile code to JavaScript file");
+		("-swf",Arg.String (set_platform Flash "flash"),"<file> : compile code to Flash SWF file");
+		("-swf9",Arg.String (fun file ->
+			set_platform Flash "flash" file;
+			if com.flash_version < 9 then com.flash_version <- 9;
+		),"<file> : compile code to Flash9 SWF file");
 		("-as3",Arg.String (fun dir ->
-			check_targets();
-			swf_infos.swf_version <- 9;
-			Plugin.define "as3gen";
-			Typer.forbidden_packages := ["js"; "neko"];
-			target := As3 dir;
+			set_platform Flash "flash" dir;
+			if com.flash_version < 9 then com.flash_version <- 9;
+			gen_as3 := true;
+			Common.define com "as3";
+			Common.define com "no_inline";
 		),"<directory> : generate AS3 code into target directory");
-		("-swf",Arg.String (fun file ->
-			check_targets();
-			Typer.forbidden_packages := ["js"; "neko"];
-			target := Swf file
-		),"<file> : compile code to Flash SWF file");
+		("-neko",Arg.String (set_platform Neko "neko"),"<file> : compile code to Neko Binary");
+		("-php",Arg.String (fun dir ->
+			classes := (["php"],"Boot") :: !classes;
+			classes := (["php"],"PhpXml__") :: !classes;
+			classes := (["php"],"PhpDate__") :: !classes;
+			classes := (["php"],"PhpMath__") :: !classes;
+			set_platform Php "php" dir;
+		),"<directory> : generate PHP code into target directory");
+		("-cpp",Arg.String (fun dir ->
+			set_platform Cpp "cpp" dir;
+		),"<directory> : generate C++ code into target directory");
+		("-xml",Arg.String (fun file ->
+			Parser.use_doc := true;
+			xml_out := Some file
+		),"<file> : generate XML types description");
+		("-main",Arg.String (fun cl ->
+			if com.main_class <> None then raise (Arg.Bad "Multiple -main");
+			let cpath = make_path cl in
+			com.main_class <- Some cpath;
+			classes := cpath :: !classes
+		),"<class> : select startup class");
+		("-lib",Arg.String (fun l ->
+			libs := l :: !libs;
+			Common.define com l;
+		),"<library[:version]> : use an haxelib library");
+		("-D",Arg.String (fun var ->
+			if var = "use_rtti_doc" then Parser.use_doc := true;
+			Common.define com var
+		),"<var> : define a conditional compilation flag");
+		("-v",Arg.Unit (fun () ->
+			if not !display then com.verbose <- true
+		),": turn on verbose mode");
+		("-debug", Arg.Unit (fun() ->
+			Common.define com "debug"; com.debug <- true)
+		, ": add debug informations to the compiled code");
+	] in
+	let adv_args_spec = [
 		("-swf-version",Arg.Int (fun v ->
-			swf_infos.swf_version <- v;
-		),"<version> : change the SWF version (6,7,8,9)");
+			com.flash_version <- v;
+		),"<version> : change the SWF version (6 to 10)");
 		("-swf-header",Arg.String (fun h ->
 			try
-				swf_infos.swf_header <- Some (match ExtString.String.nsplit h ":" with
+				swf_header := Some (match ExtString.String.nsplit h ":" with
 				| [width; height; fps] ->
 					(int_of_string width,int_of_string height,float_of_string fps,0xFFFFFF)
 				| [width; height; fps; color] ->
@@ -267,56 +300,32 @@ try
 				_ -> raise (Arg.Bad "Invalid SWF header format")
 		),"<header> : define SWF header (width:height:fps:color)");
 		("-swf-lib",Arg.String (fun file ->
-			if swf_infos.swf_lib <> None then raise (Arg.Bad "Only one SWF Library is allowed");
-			swf_infos.swf_lib <- Some file
+			if !swf_lib <> None then raise (Arg.Bad "Only one SWF Library is allowed");
+			swf_lib := Some file
 		),"<file> : add the SWF library to the compiled SWF");
-		("-neko",Arg.String (fun file ->
-			check_targets();
-			Typer.forbidden_packages := ["js"; "flash"];
-			target := Neko file
-		),"<file> : compile code to Neko Binary");
 		("-x", Arg.String (fun file ->
-			check_targets();
-			Typer.forbidden_packages := ["js"; "flash"];
 			let neko_file = file ^ ".n" in
-			target := Neko neko_file;
-			if !main_class = None then begin
+			set_platform Neko "neko" neko_file;
+			if com.main_class = None then begin
 				let cpath = make_path file in
-				main_class := Some cpath;
+				com.main_class <- Some cpath;
 				classes := cpath :: !classes
 			end;
 			cmds := ("neko " ^ neko_file) :: !cmds;
 		),"<file> : shortcut for compiling and executing a neko file");
-		("-xml",Arg.String (fun file ->
-			Parser.use_doc := true;
-			xml_out := Some file
-		),"<file> : generate XML types description");
-		("-main",Arg.String (fun cl ->
-			if !main_class <> None then raise (Arg.Bad "Multiple -main");
-			let cpath = make_path cl in
-			main_class := Some cpath;
-			classes := cpath :: !classes
-		),"<class> : select startup class");
-		("-lib",Arg.String (fun l -> libs := l :: !libs),"<library[:version]> : use an haxelib library");
-		("-D",Arg.String (fun def ->
-			Plugin.define def;
-		),"<var> : define a conditional compilation flag");
 		("-resource",Arg.String (fun res ->
-			match ExtString.String.nsplit res "@" with
-			| [file; name] ->
-				let file = (try Plugin.find_file file with Not_found -> file) in
-				let data = Std.input_file ~bin:true file in
-				if Hashtbl.mem hres name then failwith ("Duplicate resource name " ^ name);
-				Hashtbl.add hres name data
-			| [file] ->
-				let file = (try Plugin.find_file file with Not_found -> file) in
-				let data = Std.input_file ~bin:true file in
-				Hashtbl.replace hres file data
-			| _ ->
-				raise (Arg.Bad "Invalid Resource format : should be file@name")
-		),"<file@name> : add a named resource file");
+			let file, name = (match ExtString.String.nsplit res "@" with
+				| [file; name] -> file, name
+				| [file] -> file, file
+				| _ -> raise (Arg.Bad "Invalid Resource format : should be file@name")
+			) in
+			let file = (try Common.find_file com file with Not_found -> file) in
+			let data = Std.input_file ~bin:true file in
+			if Hashtbl.mem com.resources name then failwith ("Duplicate resource name " ^ name);
+			Hashtbl.add com.resources name data
+		),"<file>[@name] : add a named resource file");
 		("-exclude",Arg.String (fun file ->
-			let file = (try Plugin.find_file file with Not_found -> file) in
+			let file = (try Common.find_file com file with Not_found -> file) in
 			let ch = open_in file in
 			let lines = Std.input_list ch in
 			close_in ch;
@@ -328,40 +337,43 @@ try
 				| x :: l -> (List.rev l,x)
 			) lines) @ !excludes;
 		),"<filename> : don't generate code for classes listed in this file");
-		("-v",Arg.Unit (fun () -> Plugin.verbose := true),": turn on verbose mode");
-		("-debug", define "debug", ": add debug informations to the compiled code");
 		("-prompt", Arg.Unit (fun() -> prompt := true),": prompt on error");
-		("-cmd", Arg.String (fun cmd ->
-			cmds := cmd :: !cmds
+		("-cmd", Arg.String (fun cmd ->			
+			cmds := expand_env cmd :: !cmds
 		),": run the specified command after successful compilation");
 		("--flash-strict", define "flash_strict", ": more type strict flash API");
-		("--override", Arg.Unit (fun() ->
-			Typer.check_override := true
-		),": ensure that overriden methods are declared with 'override'");
 		("--no-traces", define "no_traces", ": don't compile trace calls in the program");
 		("--flash-use-stage", define "flash_use_stage", ": place objects found on the stage of the SWF lib");
 		("--neko-source", define "neko_source", ": keep generated neko source");
 		("--gen-hx-classes", Arg.String (fun file ->
-			gen_hx := true;
-			Genas3.genhx file
+			com.file <- file;
+			Genas3.genhx com;
+			did_something := true;
 		),"<file> : generate hx headers from SWF9 file");
 		("--next", Arg.Unit (fun() -> assert false), ": separate several haxe compilations");
-		("--altfmt", Arg.Unit (fun() -> alt_format := true),": use alternative error output format");
-		("--auto-xml", Arg.Unit (fun() -> auto_xml := true),": automatically create an XML for each target");
 		("--display", Arg.String (fun file_pos ->
 			let file, pos = try ExtString.String.split file_pos "@" with _ -> failwith ("Invalid format : " ^ file_pos) in
 			let pos = try int_of_string pos with _ -> failwith ("Invalid format : "  ^ pos) in
 			display := true;
 			no_output := true;
 			Parser.resume_display := {
-				Ast.pfile = (!Plugin.get_full_path) file;
+				Ast.pfile = Common.get_full_path file;
 				Ast.pmin = pos;
 				Ast.pmax = pos;
 			};
 		),": display code tips");
 		("--no-output", Arg.Unit (fun() -> no_output := true),": compiles but does not generate any file");
-		("--times", Arg.Unit (fun() -> Plugin.times := true),": mesure compilation times");
-		("--no-inline", Arg.Unit (fun() -> Plugin.define "no_inline"),": disable inlining");
+		("--times", Arg.Unit (fun() -> measure_times := true),": mesure compilation times");
+		("--no-inline", define "no_inline", ": disable inlining");
+		("--no-opt", define "no_opt", ": disable code optimizations");		
+		("--php-front",Arg.String (fun f ->
+			if com.php_front <> None then raise (Arg.Bad "Multiple --php-front");
+			com.php_front <- Some f;
+		),"<filename> : select the name for the php front file");
+		("--remap", Arg.String (fun s ->
+			let pack, target = (try ExtString.String.split s ":" with _ -> raise (Arg.Bad "Invalid format")) in
+			com.package_rules <- PMap.add pack (Remap target) com.package_rules;
+		),"<package:target> : remap a package to another one");
 	] in
 	let current = ref 0 in
 	let args = Array.of_list ("" :: params) in
@@ -371,13 +383,13 @@ try
 			let hxml_args = parse_hxml cl in
 			let p1 = Array.to_list (Array.sub args 1 (!current - 1)) in
 			let p2 = Array.to_list (Array.sub args (!current + 1) (Array.length args - !current - 1)) in
-			if !Plugin.verbose then print_endline ("Processing HXML : " ^ cl);
+			if com.verbose then print_endline ("Processing HXML : " ^ cl);
 			process_params [] (p1 @ hxml_args @ p2);
 			raise Hxml_found
 		| _ ->
 			classes := make_path cl :: !classes
 	in
-	Arg.parse_argv ~current args args_spec args_callback usage;
+	Arg.parse_argv ~current args (basic_args_spec @ adv_args_spec) args_callback usage;
 	(match !libs with
 	| [] -> ()
 	| l ->
@@ -396,67 +408,82 @@ try
 				l :: acc
 		) [] lines in
 		if ret <> Unix.WEXITED 0 then failwith (String.concat "\n" lines);
-		Plugin.class_path := lines @ !Plugin.class_path;
-	);
-	(match !target with
-	| No ->
-		()
-	| Swf file | As3 file ->
-		(* check file extension. In case of wrong commandline, we don't want
-		   to accidentaly delete a source file. *)
-		if not !no_output && file_extension file = "swf" then delete_file file;
-		Plugin.define "flash";
-		Plugin.define ("flash"  ^ string_of_int swf_infos.swf_version);
-	| Neko file ->
-		if not !no_output && file_extension file = "n" then delete_file file;
-		Plugin.define "neko";
-	| Js file ->
-		if not !no_output && file_extension file = "js" then delete_file file;
-		Plugin.define "js";
+		com.class_path <- lines @ com.class_path;
 	);
+	let ext = (match com.platform with
+		| Cross ->
+			(* no platform selected *)
+			set_platform Cross "cross" "";
+			no_output := true; ""
+		| Flash | Flash9 ->
+			Common.define com ("flash" ^ string_of_int com.flash_version);
+			if com.flash_version >= 9 then begin
+				Common.define com "flash9"; (* always define flash9, even for flash10+ *)
+				com.package_rules <- PMap.add "flash" (Directory "flash9") com.package_rules;
+				com.package_rules <- PMap.add "flash9" (Directory "flash9") com.package_rules;
+				com.platform <- Flash9;
+			end;
+			"swf"
+		| Neko -> "n"
+		| Js -> "js"
+		| Php -> "php"
+		| Cpp -> "cpp"
+	) in
+	(* check file extension. In case of wrong commandline, we don't want
+		to accidentaly delete a source file. *)
+	if not !no_output && file_extension com.file = ext then delete_file com.file;
 	if !classes = [([],"Std")] then begin
-		if !cmds = [] && not !gen_hx then Arg.usage args_spec usage;
+		if !cmds = [] && not !did_something then Arg.usage basic_args_spec usage;
 	end else begin
-		if !Plugin.verbose then print_endline ("Classpath : " ^ (String.concat ";" !Plugin.class_path));
-		let t = Plugin.timer "typing" in
-		let ctx = Typer.context type_error warn in
-		List.iter (fun cpath -> ignore(Typer.load ctx cpath Ast.null_pos)) (List.rev !classes);
+		if com.verbose then print_endline ("Classpath : " ^ (String.concat ";" com.class_path));
+		let t = Common.timer "typing" in
+		Typecore.type_expr_ref := (fun ctx e need_val -> Typer.type_expr ~need_val ctx e);
+		Typecore.build_inheritance := Codegen.on_inherit;
+		let ctx = Typer.create com in
+		List.iter (fun cpath -> ignore(com.type_api.load_module cpath Ast.null_pos)) (List.rev !classes);
 		Typer.finalize ctx;
 		t();
 		if !has_error then do_exit();
-		if !display then begin
-			xml_out := None;
-			auto_xml := false;
-		end;
-		if !no_output then target := No;
-		let do_auto_xml file = if !auto_xml then xml_out := Some (file ^ ".xml") in
-		let types = Typer.types ctx (!main_class) (!excludes) in
-		(match !target with
-		| No -> ()
-		| Swf file ->
-			do_auto_xml file;
-			if !Plugin.verbose then print_endline ("Generating swf : " ^ file);
-			Genswf.generate file swf_infos types hres
-		| Neko file ->
-			do_auto_xml file;
-			if !Plugin.verbose then print_endline ("Generating neko : " ^ file);
-			Genneko.generate file types hres !libs
-		| Js file ->
-			do_auto_xml file;
-			if !Plugin.verbose then print_endline ("Generating js : " ^ file);
-			Genjs.generate file types hres
-		| As3 dir ->
-			if !Plugin.verbose then print_endline ("Generating AS3 in : " ^ dir);
-			Genas3.generate dir types
+		if !display then xml_out := None;
+		if !no_output then com.platform <- Cross;
+		com.types <- Typer.types ctx com.main_class (!excludes);
+		com.lines <- Lexer.build_line_index();
+		let filters = [
+			Codegen.check_local_vars_init;
+			Codegen.block_vars com;
+		] in
+		let filters = (if Common.defined com "no_opt" then filters else Optimizer.reduce_expression com :: filters) in
+		Codegen.post_process com filters;
+		(match com.platform with
+		| Cross ->
+			()
+		| Flash | Flash9 when !gen_as3 ->
+			if com.verbose then print_endline ("Generating AS3 in : " ^ com.file);
+			Genas3.generate com;
+		| Flash | Flash9 ->
+			if com.verbose then print_endline ("Generating swf : " ^ com.file);
+			Genswf.generate com !swf_header !swf_lib;
+		| Neko ->
+			if com.verbose then print_endline ("Generating neko : " ^ com.file);
+			Genneko.generate com !libs;
+		| Js ->
+			if com.verbose then print_endline ("Generating js : " ^ com.file);
+			Genjs.generate com
+		| Php ->
+			if com.verbose then print_endline ("Generating PHP in : " ^ com.file);
+			Genphp.generate com;
+		| Cpp ->
+			if com.verbose then print_endline ("Generating Cpp in : " ^ com.file);
+			Gencpp.generate com;
 		);
 		(match !xml_out with
 		| None -> ()
 		| Some file ->
-			if !Plugin.verbose then print_endline ("Generating xml : " ^ file);
-			Genxml.generate file ctx types);
+			if com.verbose then print_endline ("Generating xml : " ^ com.file);
+			Genxml.generate com file);
 	end;
 	if not !no_output then List.iter (fun cmd ->
-		let t = Plugin.timer "command" in
+		let t = Common.timer "command" in
 		let len = String.length cmd in
 		if len > 3 && String.sub cmd 0 3 = "cd " then
 			Sys.chdir (String.sub cmd 3 (len - 3))
@@ -465,13 +492,18 @@ try
 		t();
 	) (List.rev !cmds)
 with
+	| Common.Abort (m,p) -> report m p
 	| Lexer.Error (m,p) -> report (Lexer.error_msg m) p
 	| Parser.Error (m,p) -> report (Parser.error_msg m) p
-	| Typer.Error (m,p) -> report (Typer.error_msg m) p
+	| Typecore.Error (m,p) -> report (Typecore.error_msg m) p
 	| Failure msg | Arg.Bad msg -> report ("Error : " ^ msg) Ast.null_pos
 	| Arg.Help msg -> print_string msg
 	| Hxml_found -> ()
 	| Typer.Display t ->
+		(*
+			documentation is currently not output even when activated
+			because the parse 'eats' it when used in "resume" mode
+		*)
 		let ctx = Type.print_context() in
 		(match Type.follow t with
 		| Type.TAnon a ->
@@ -487,7 +519,7 @@ with
 			prerr_endline "</type>");
 		exit 0;
 	| Parser.TypePath p ->
-		let packs, classes = read_type_path p (!Plugin.class_path) in
+		let packs, classes = read_type_path com p in
 		if packs = [] && classes = [] then report ("No classes found in " ^ String.concat "." p) Ast.null_pos;
 		report_list (List.map (fun f -> f,"","") (packs @ classes));
 		exit 0;
@@ -495,17 +527,15 @@ with
 		report (Printexc.to_string e) Ast.null_pos
 
 ;;
-let all = Plugin.timer "other" in
-Plugin.times := false;
-Plugin.get_full_path := get_full_path;
+let all = Common.timer "other" in
 process_params [] (List.tl (Array.to_list Sys.argv));
 all();
-if !Plugin.times then begin
+if !measure_times then begin
 	let tot = ref 0. in
-	Hashtbl.iter (fun _ t -> tot := !tot +. t.Plugin.total) Plugin.htimers;
+	Hashtbl.iter (fun _ t -> tot := !tot +. t.total) Common.htimers;
 	Printf.eprintf "Total time : %.3fs\n" !tot;
 	Printf.eprintf "------------------------------------\n";
 	Hashtbl.iter (fun _ t ->
-		Printf.eprintf "  %s : %.3fs, %.0f%%\n" t.Plugin.name t.Plugin.total (t.Plugin.total *. 100. /. !tot);
-	) Plugin.htimers;
+		Printf.eprintf "  %s : %.3fs, %.0f%%\n" t.name t.total (t.total *. 100. /. !tot);
+	) Common.htimers;
 end;
\ No newline at end of file
