@@ -103,6 +103,19 @@ let ops , ops_ids =
 		A3OIAdd, 0xC5;
 		A3OISub, 0xC6;
 		A3OIMul, 0xC7;
+		A3OMemGet8, 0x35;
+		A3OMemGet16, 0x36;
+		A3OMemGet32, 0x37;
+		A3OMemGetFloat, 0x38;
+		A3OMemGetDouble, 0x39;
+		A3OMemSet8, 0x3A;
+		A3OMemSet16, 0x3B;
+		A3OMemSet32, 0x3C;
+		A3OMemSetFloat, 0x3D;
+		A3OMemSetDouble, 0x3E;
+		A3OSign1, 0x50;
+		A3OSign8, 0x51;
+		A3OSign16, 0x52;
 	];
 	h , h2
 
@@ -121,6 +134,7 @@ let length = function
 	| A3CallStack n
 	| A3ConstructSuper n
 	| A3BreakPointLine n
+	| A3ApplyType n
 	| A3DebugLine n ->
 		1 + int_length n
 	| A3GetSlot s
@@ -128,6 +142,7 @@ let length = function
 		1 + int_length s
 	| A3ClassDef n ->
 		1 + int_length (int_index_nz n)
+	| A3DxNs f
 	| A3String f
 	| A3DebugFile f ->
 		1 + int_length (int_index f)
@@ -151,6 +166,7 @@ let length = function
 	| A3SetProp f
 	| A3Cast f
 	| A3GetSuper f
+	| A3GetDescendants f
 	| A3SetSuper f ->
 		1 + int_length (int_index f)
 	| A3Op _
@@ -192,6 +208,7 @@ let length = function
 	| A3HasNext
 	| A3SetThis
 	| A3Timestamp
+	| A3DxNsLate
 	| A3Unk _ -> 1
 	| A3AsType n | A3IsType n ->
 		1 + int_length (int_index n)
@@ -222,8 +239,8 @@ let opcode ch =
 	| 0x03 -> A3Throw
 	| 0x04 -> A3GetSuper (read_index ch)
 	| 0x05 -> A3SetSuper (read_index ch)
-	(* 0x06 -> E4X *)
-	(* 0x07 -> E4X *)
+	| 0x06 -> A3DxNs (read_index ch)
+	| 0x07 -> A3DxNsLate
 	| 0x08 -> A3RegKill (read_int ch)
 	| 0x09 -> A3Label
 	(* 0x0A -> NONE *)
@@ -321,12 +338,14 @@ let opcode ch =
 		let id = read_index ch in
 		let nargs = read_int ch in
 		A3CallPropVoid (id,nargs)
-	(* 0x50 - 0x54 -> NONE *)
+	(* 0x50 - 0x52 -> NONE *)
+	| 0x53 -> A3ApplyType (read_int ch)
+	(* 0x54 -> NONE *)
 	| 0x55 -> A3Object (read_int ch)
 	| 0x56 -> A3Array (read_int ch)
 	| 0x57 -> A3NewBlock
 	| 0x58 -> A3ClassDef (read_index_nz ch)
-	(* 0x59 -> E4X *)
+	| 0x59 -> A3GetDescendants (read_index ch)
 	| 0x5A -> A3Catch (read_int ch)
 	(* 0x5B -> NONE *)
 	(* 0x5C -> NONE *)
@@ -441,6 +460,11 @@ let write ch = function
 	| A3SetSuper f ->
 		write_byte ch 0x05;
 		write_index ch f
+	| A3DxNs i ->
+		write_byte ch 0x06;
+		write_index ch i
+	| A3DxNsLate ->
+		write_byte ch 0x07
 	| A3RegKill n ->
 		write_byte ch 0x08;
 		write_int ch n
@@ -571,6 +595,9 @@ let write ch = function
 		write_byte ch 0x4F;
 		write_index ch f;
 		write_int ch n
+	| A3ApplyType n ->
+		write_byte ch 0x53;
+		write_int ch n
 	| A3Object n ->
 		write_byte ch 0x55;
 		write_int ch n
@@ -582,6 +609,9 @@ let write ch = function
 	| A3ClassDef f ->
 		write_byte ch 0x58;
 		write_index_nz ch f
+	| A3GetDescendants f ->
+		write_byte ch 0x59;
+		write_index ch f
 	| A3Catch n ->
 		write_byte ch 0x5A;
 		write_int ch n
@@ -741,6 +771,19 @@ let dump_op = function
 	| A3OIAdd -> "iadd"
 	| A3OISub -> "isub"
 	| A3OIMul -> "imul"
+	| A3OMemSet8 -> "mset8"
+	| A3OMemSet16 -> "set16"
+	| A3OMemSet32 -> "mset32"
+	| A3OMemSetFloat -> "msetfloat"
+	| A3OMemSetDouble -> "msetdouble"
+	| A3OMemGet8 -> "mget8"
+	| A3OMemGet16 -> "mget16"
+	| A3OMemGet32 -> "mget32"
+	| A3OMemGetFloat -> "mgetfloat"
+	| A3OMemGetDouble -> "mgetdouble"
+	| A3OSign1 -> "sign1"
+	| A3OSign8 -> "sign8"
+	| A3OSign16 -> "sign16"
 
 let dump_jump = function
 	| J3NotLt -> "-nlt"
@@ -761,12 +804,13 @@ let dump_jump = function
 
 let dump ctx op =
 	let ident n = ctx.as3_idents.(int_index n - 1) in
-	let field n =
+	let rec field n =
 		let t = ctx.as3_names.(int_index n - 1) in
 		match t with
 		| A3MMultiName (Some ident,_) -> "[" ^ iget ctx.as3_idents ident ^ "]"
 		| A3MName (ident,_) -> iget ctx.as3_idents ident
 		| A3MMultiNameLate idx -> "~array"
+		| A3MParams (t,params) -> field t ^ "<" ^ String.concat "." (List.map field params) ^ ">"
 		| _ -> "???"
 	in
 	match op with
@@ -775,6 +819,8 @@ let dump ctx op =
 	| A3Throw -> "throw"
 	| A3GetSuper f -> s "getsuper %s" (field f)
 	| A3SetSuper f -> s "setsuper %s" (field f)
+	| A3DxNs i -> s "dxns %s" (ident i)
+	| A3DxNsLate -> "dxnslate"
 	| A3RegKill n -> s "kill %d" n
 	| A3Label -> "label"
 	| A3Jump (k,n) -> s "jump%s %d" (dump_jump k) n
@@ -815,10 +861,12 @@ let dump ctx op =
 	| A3CallPropLex (f,n) -> s "callproplex %s (%d)" (field f) n
 	| A3CallSuperVoid (f,n) -> s "callsupervoid %s (%d)" (field f) n
 	| A3CallPropVoid (f,n) -> s "callpropvoid %s (%d)" (field f) n
+	| A3ApplyType n -> s "applytype %d" n
 	| A3Object n -> s "object %d" n
 	| A3Array n -> s "array %d" n
 	| A3NewBlock -> "newblock"
 	| A3ClassDef n -> s "classdef %d" (int_index_nz n)
+	| A3GetDescendants f -> s "getdescendants %s" (field f)
 	| A3Catch n -> s "catch %d" n
 	| A3FindPropStrict f -> s "findpropstrict %s" (field f)
 	| A3FindProp f -> s "findprop %s" (field f)
