@@ -86,6 +86,9 @@ let as3_int_length i =
 	else
 		1
 
+let as3_uint_length i =
+	as3_int_length i
+
 let sum f l =
 	List.fold_left (fun acc n -> acc + f n) 0 l
 
@@ -136,6 +139,8 @@ let rec as3_name_length t =
 		idx_length idx
 	| A3MAttrib n ->
 		as3_name_length n - 1
+	| A3MParams (id,pl) ->
+		idx_length id + 1 + (sum idx_length pl)
 
 let as3_value_length extra = function
 	| A3VNone -> if extra then 2 else 1
@@ -222,7 +227,7 @@ let as3_length ctx =
 	String.length ctx.as3_unknown +
 	4 +
 	list_length as3_int_length ctx.as3_ints +
-	1 +
+	list_length as3_uint_length ctx.as3_uints +
 	list_length (fun _ -> 8) ctx.as3_floats
 	+ if parse_idents then list_length as3_ident_length ctx.as3_idents
 	+ if parse_namespaces then list_length (as3_namespace_length ei) ctx.as3_namespaces
@@ -267,7 +272,7 @@ let read_as3_int ch =
 	Int32.logor big small
 
 let read_as3_uint ch =
-	assert false
+	read_as3_int ch
 
 let read_int ch =
 	Int32.to_int (read_as3_int ch)
@@ -348,7 +353,18 @@ let rec read_name ctx ?k ch =
 		A3MMultiNameLate ns
 	| 0x1C ->
 		A3MAttrib (read_name ctx ~k:0x1B ch)
+	| 0x1D ->
+		let rec loop n =
+			if n = 0 then
+				[]
+			else
+				let name = magic_index (read_int ch) in
+				name :: loop (n - 1)
+		in
+		let id = magic_index (read_int ch) in
+		A3MParams (id,loop (IO.read_byte ch))
 	| n ->
+		prerr_endline (string_of_int n);
 		assert false
 
 let read_value ctx ch extra =
@@ -610,6 +626,8 @@ let write_as3_int ch i =
 	end else
 		IO.write_byte ch a
 
+let write_as3_uint = write_as3_int
+
 let write_int ch i =
 	write_as3_int ch (Int32.of_int i)
 
@@ -688,8 +706,13 @@ let rec write_name ch ?k x =
 			| A3MRuntimeName _ -> 0x10
 			| A3MRuntimeNameLate -> 0x12
 			| A3MMultiNameLate _ -> 0x1C
-			| A3MAttrib _ -> assert false
+			| A3MAttrib _ | A3MParams _ -> assert false
 		) n
+	| A3MParams (id,pl) ->
+		IO.write_byte ch (b 0x1D);
+		write_index ch id;
+		IO.write_byte ch (List.length pl);
+		List.iter (write_index ch) pl
 
 let write_value ch extra v =
 	match v with
@@ -835,7 +858,7 @@ let write ch1 ctx =
 	let empty_index = as3_empty_index ctx in
 	IO.write_i32 ch header_magic;
 	write_list ch write_as3_int ctx.as3_ints;
-	IO.write_byte ch 0;
+	write_list ch write_as3_uint ctx.as3_uints;
 	write_list ch IO.write_double ctx.as3_floats;
 	if parse_idents then write_list ch write_as3_ident ctx.as3_idents;
 	if parse_namespaces then write_list ch (write_namespace empty_index) ctx.as3_namespaces;
@@ -854,6 +877,8 @@ let write ch1 ctx =
 (* ************************************************************************ *)
 (* DUMP *)
 
+let dump_code_size = ref true
+
 let ident_str ctx i =
 	iget ctx.as3_idents i
 
@@ -875,7 +900,7 @@ let ns_set_str ctx i =
 	let l = iget ctx.as3_nsets i in
 	String.concat " " (List.map (fun r -> namespace_str ctx r) l)
 
-let name_str ctx kind t =
+let rec name_str ctx kind t =
 	let rec loop = function
 		| A3MName (id,r) -> Printf.sprintf "%s %s%s" (namespace_str ctx r) kind (ident_str ctx id)
 		| A3MMultiName (id,r) -> Printf.sprintf "[%s %s%s]" (ns_set_str ctx r) kind (match id with None -> "NO" | Some i -> ident_str ctx i)
@@ -883,6 +908,7 @@ let name_str ctx kind t =
 		| A3MRuntimeNameLate -> "RTLATE"
 		| A3MMultiNameLate id -> Printf.sprintf "late:(%s)" (ns_set_str ctx id)
 		| A3MAttrib n -> "attrib " ^ loop n
+		| A3MParams (id,pl) -> Printf.sprintf "%s<%s>" (name_str ctx kind id) (String.concat "," (List.map (name_str ctx kind) pl))
 	in
 	loop (iget ctx.as3_names t)
 
@@ -1008,7 +1034,7 @@ let dump_function ctx ch idx f =
 	let pos = ref 0 in
 	Array.iter (fun op ->
 		IO.printf ch "%4d    %s\n" !pos (As3code.dump ctx op);
-		pos := !pos + As3code.length op;
+		if !dump_code_size then pos := !pos + As3code.length op else incr pos;
 	) f.fun3_code;
 	IO.printf ch "\n"
 
@@ -1040,7 +1066,7 @@ let dump ch ctx id =
 	(match id with
 	| None -> IO.printf ch "\n---------------- AS3 -------------------------\n\n";
 	| Some (id,f) -> IO.printf ch "\n---------------- AS3 %s [%d] -----------------\n\n" f id);
-	Array.iteri (dump_int ctx ch) ctx.as3_ints;
+(*	Array.iteri (dump_int ctx ch) ctx.as3_ints;
 	Array.iteri (dump_float ctx ch) ctx.as3_floats;
 	Array.iteri (dump_ident ctx ch) ctx.as3_idents;
 	IO.printf ch "\n";
@@ -1049,7 +1075,7 @@ let dump ch ctx id =
 	Array.iteri (dump_ns_set ctx ch) ctx.as3_nsets;
 	IO.printf ch "\n";
 	Array.iteri (dump_name ctx ch) ctx.as3_names;
-	IO.printf ch "\n";
+	IO.printf ch "\n"; *)
 (*	Array.iteri (dump_metadata ctx ch) ctx.as3_metadatas; *)
 	Array.iteri (dump_class ctx ch) ctx.as3_classes;
 	Array.iteri (dump_init ctx ch) ctx.as3_inits;
